//-----------------------------------------------------------------------------
// name: VisualSine.cpp
// desc: ThePrinciple - MUSIC 256A hw1
//
// author: Jingjie Zhang(jingjiez@ccrma.stanford.edu)
//   date: fall 2018
//  based: VisualSine template by Ge Wang
//   uses: RtAudio by Gary Scavone
//-----------------------------------------------------------------------------
#include "RtAudio/RtAudio.h"
#include "chuck.h"
#include <math.h>
#include <stdlib.h>
#include <iostream>
using namespace std;

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "chuck_fft.h" // FFT
#define PI 3.14159265f

//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void initGfx();
void idleFunc();
void displayFunc();
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );

// our datetype
#define SAMPLE float
// corresponding format for RtAudio
#define MY_FORMAT RTAUDIO_FLOAT32
// sample rate
#define MY_SRATE 44100
// number of channels
#define MY_CHANNELS 2
// for convenience
#define MY_PIE 3.14159265358979

// width and height
long g_width = 1024;
long g_height = 720;

// global buffer
SAMPLE* g_buffer = NULL;
SAMPLE* g_window = NULL;
long g_bufferSize;
long g_fftSize;

int g_waterfall_depth = 64;
int g_waterfall_ptr = 0;
float** g_waterfall = new float*[g_waterfall_depth];

//whether the narrative is enabled
bool g_narrative_enabled = false;
// global chuck instance
ChucK * the_chuck;

//-----------------------------------------------------------------------------
// name: help()
// desc: ...
//-----------------------------------------------------------------------------
void help()
{
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "ThePrinciple - MUSIC 256A hw1\n" );
    fprintf( stderr, "Jingjie Zhang\n" );
    fprintf( stderr, "http://ccrma.stanford.edu/~jingjiez\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "Press 'q' - quit the program\n" );
    fprintf( stderr, "Press SPACE - play narrative\n" );
    fprintf( stderr, "\n" );
}

//-----------------------------------------------------------------------------
// name: callme()
// desc: audio callback
//-----------------------------------------------------------------------------
int callme( void * outputBuffer, void * inputBuffer, unsigned int numFrames,
            double streamTime, RtAudioStreamStatus status, void * data )
{
    // cast!
    SAMPLE * input = (SAMPLE *)inputBuffer;
    SAMPLE * output = (SAMPLE *)outputBuffer;

    // compute chuck!
    if (g_narrative_enabled) the_chuck->run(input, output, numFrames);

    // fill
    for( int i = 0; i < numFrames; ++i ) {
        g_buffer[i] = input[i*MY_CHANNELS] + output[i*MY_CHANNELS];
        if (!g_narrative_enabled)
            for( int j = 0; j < MY_CHANNELS; j++ ) {
                // mute output if the narrative is not running
                output[i*MY_CHANNELS + j] = 0;
            }
    }

    return 0;
}


//-----------------------------------------------------------------------------
// name: main()
// desc: entry point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    // print usage
    help();

    // instantiate RtAudio object
    RtAudio audio;
    // variables
    unsigned int bufferBytes = 0;
    // frame size
    unsigned int bufferFrames = 1024;

    // check for audio devices
    if( audio.getDeviceCount() < 1 )
    {
        // nopes
        cout << "no audio devices found!" << endl;
        exit( 1 );
    }

    // initialize GLUT
    glutInit( &argc, argv );
    // init gfx
    initGfx();

    // let RtAudio print messages to stderr.
    audio.showWarnings( true );

    // set input and output parameters
    RtAudio::StreamParameters iParams, oParams;
    iParams.deviceId = audio.getDefaultInputDevice();
    iParams.nChannels = MY_CHANNELS;
    iParams.firstChannel = 0;
    oParams.deviceId = audio.getDefaultOutputDevice();
    oParams.nChannels = MY_CHANNELS;
    oParams.firstChannel = 0;

    // create stream options
    RtAudio::StreamOptions options;

    // go for it
    try {
        // open a stream
        audio.openStream( &oParams, &iParams, MY_FORMAT, MY_SRATE, &bufferFrames, &callme, (void *)&bufferBytes, &options );
    }
    catch( RtError& e )
    {
        // error!
        cout << e.getMessage() << endl;
        exit( 1 );
    }

    // compute
    bufferBytes = bufferFrames * MY_CHANNELS * sizeof(SAMPLE);
    // allocate global buffer
    g_bufferSize = bufferFrames;
    g_fftSize = bufferFrames * 2;
    g_buffer = new SAMPLE[g_bufferSize];
    memset( g_buffer, 0, sizeof(SAMPLE)*g_bufferSize );
    g_window = new SAMPLE[g_bufferSize];
    hanning(g_window, g_bufferSize);
    for (int i = 0; i < g_waterfall_depth; ++i) {
        g_waterfall[i] = new float[bufferFrames];
    }

    // set up chuck
    the_chuck = new ChucK();
    // set sample rate and number of in/out channels on our chuck
    the_chuck->setParam(CHUCK_PARAM_SAMPLE_RATE, MY_SRATE);
    the_chuck->setParam(CHUCK_PARAM_INPUT_CHANNELS, MY_CHANNELS);
    the_chuck->setParam(CHUCK_PARAM_OUTPUT_CHANNELS, MY_CHANNELS);
    the_chuck->setParam(CHUCK_PARAM_WORKING_DIRECTORY, ".");
    // initialize chuck
    the_chuck->init();
    // run a chuck program
    the_chuck->compileFile("ThePrinciple.ck", "");

    // go for it
    try {
        // start stream
        audio.startStream();

        // let GLUT handle the current thread from here
        glutMainLoop();

        // stop the stream.
        audio.stopStream();
    }
    catch( RtError& e )
    {
        // print error message
        cout << e.getMessage() << endl;
        goto cleanup;
    }

cleanup:
    // close if open
    if( audio.isStreamOpen() )
        audio.closeStream();

    // done
    return 0;
}

//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void initGfx()
{
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "ThePrinciple" );

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );

    // set clear color
    glClearColor( 0, 0, 0, 1 );
    // enable color material
    glEnable( GL_COLOR_MATERIAL );
    // enable depth test
    glEnable( GL_DEPTH_TEST );
}

//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 0.0f, 10.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}

int g_state = 3;
//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
        case 'Q':
        case 'q':
            exit(1);
            break;
        case ' ':
            if (!g_narrative_enabled) {
                g_state = 0;
                the_chuck->compileFile("ThePrinciple.ck", "");
                g_narrative_enabled = true;
            }
            else g_narrative_enabled = false;
            break;
        case 's':
            g_state += 1;
            break;
    }

    glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Draw Circles & Arcs
//-----------------------------------------------------------------------------
void drawCircle(float cx, float cy, float r, int num_segments) {
	float theta = 2.0 * PI / float(num_segments);
	float tangetial_factor = tanf(theta);//calculate the tangential factor
	float radial_factor = cosf(theta);//calculate the radial factor
	float x = r;//we start at angle = 0
	float y = 0;
	glBegin(GL_LINE_LOOP);
	for(int ii = 0; ii < num_segments; ii++){
		glVertex2f(x + cx, y + cy);//output vertex
		//calculate the tangential vector
		float tx = -y;
		float ty = x;
		//add the tangential vector
		x += tx * tangetial_factor;
		y += ty * tangetial_factor;
		//correct using the radial factor
		x *= radial_factor;
		y *= radial_factor;
	}
	glEnd();
}
void drawArc(float cx, float cy, float r, float start_angle, float arc_angle, int num_segments) {
    //theta is now calculated from the arc angle instead, the -1 bit comes from the fact that the arc is open
	float theta = arc_angle / float(num_segments - 1);
	float tangetial_factor = tanf(theta);
	float radial_factor = cosf(theta);
	float x = r * cosf(start_angle);//we now start at the start angle
	float y = r * sinf(start_angle);
	glBegin(GL_LINE_STRIP);//since the arc is not a closed curve, this is a strip now
	for(int ii = 0; ii < num_segments; ii++) {
		glVertex2f(x + cx, y + cy);
		float tx = -y;
		float ty = x;
		x += tx * tangetial_factor;
		y += ty * tangetial_factor;
		x *= radial_factor;
		y *= radial_factor;
	}
	glEnd();
}
void drawSolidCircle(float cx, float cy, float r) {
    glEnable(GL_LINE_SMOOTH);
    glBegin(GL_TRIANGLE_FAN);
    glVertex2f(cx, cy); // Center
    for(int i = 0; i <= 360; ++i)
            glVertex2f(r*cos(M_PI * i / 180.0) + cx, r*sin(M_PI * i / 180.0) + cy);
    glEnd();
}

//-----------------------------------------------------------------------------
// Draw Rectangles
//-----------------------------------------------------------------------------
void drawBar(float cx, float cy, float length, float width) {
    glBegin(GL_QUADS);
    glVertex2f(cx - length/2, cy + width/2);
    glVertex2f(cx - length/2, cy - width/2);
    glVertex2f(cx + length/2, cy - width/2);
    glVertex2f(cx + length/2, cy + width/2);
    glEnd();
}
void drawDashed(float cx, float cy, float length, float width) {
    glBegin(GL_QUADS);
    glVertex2f(cx - length/2, cy + width/2);
    glVertex2f(cx - length/2, cy - width/2);
    glVertex2f(cx - length/10, cy - width/2);
    glVertex2f(cx - length/10, cy + width/2);
    glEnd();
    glBegin(GL_QUADS);
    glVertex2f(cx + length/10, cy + width/2);
    glVertex2f(cx + length/10, cy - width/2);
    glVertex2f(cx + length/2, cy - width/2);
    glVertex2f(cx + length/2, cy + width/2);
    glEnd();
}

//-----------------------------------------------------------------------------
// Find Max Sample in Buffer
//-----------------------------------------------------------------------------
SAMPLE findMaxSample(SAMPLE* buffer) {
    float current_max_sample = buffer[0];
    for (int i = 1; i < g_bufferSize; ++i) {
        if (buffer[i] >= current_max_sample) current_max_sample = buffer[i];
    }
    return current_max_sample;
}

//-----------------------------------------------------------------------------
// Calculate RMS
//-----------------------------------------------------------------------------
float calculateRMS() {
    float sum_sqr = 0;
    for (int i = 0; i < g_bufferSize; ++i) {
        sum_sqr += g_buffer[i];
    }
    return sqrtf(sum_sqr / g_bufferSize);
}


float g_RMS = 0;
const float g_RMS_threshold = 0.5;
const int g_state_switch_max_delay = 300;
int g_state_switch_delay_counter = 0;
GLfloat theAngle = 0;
float g_radius0 = 0;
const float g_radius = 2.4;
bool g_eye_switch = false;
int g_eye_counter = 0;
//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    //===============================================================================
    // My Visualizer
    //===============================================================================
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    // line width
    glLineWidth(1.0);
    //===============================================================================
    // Calculate RMS
    float RMS = calculateRMS();
    if (!isnan(RMS)) g_RMS = RMS;
    //===============================================================================
    // Calculate FFT
    SAMPLE buffer[g_bufferSize];
    SAMPLE fft_buffer[g_fftSize];
    apply_window((float*)g_buffer, g_window, g_bufferSize);
    memcpy(buffer, g_buffer, g_bufferSize * sizeof(SAMPLE));
    memset(fft_buffer, 0, g_fftSize * sizeof(SAMPLE));
    memcpy(fft_buffer, buffer, g_bufferSize * sizeof(SAMPLE));
    rfft((float *)fft_buffer, g_bufferSize, FFT_FORWARD);
    complex* cbuf = (complex *)fft_buffer;
    g_waterfall_ptr = (g_waterfall_ptr + 1) % g_waterfall_depth;
    for (int i = 0; i < g_bufferSize; ++i) {
        g_waterfall[g_waterfall_ptr][i] = sqrtf(300*cmp_abs(cbuf[i]));
    }
    //===============================================================================
    // A simple state machine
    //if (g_state == 0) cout << g_waterfall[g_waterfall_ptr][366] + g_waterfall[g_waterfall_ptr][367] << endl;
    //if (g_state >= 1) cout << g_waterfall[g_waterfall_ptr][1] << endl;
    //if (g_state >= 2) cout << g_waterfall[g_waterfall_ptr][1] << endl;
    if (g_state_switch_delay_counter) --g_state_switch_delay_counter;
    if (g_state == 0 && g_waterfall[g_waterfall_ptr][366] + g_waterfall[g_waterfall_ptr][367] > 1.0) {
        //cout << "=============== State 1 ===============" << endl;
        g_state = 1;
        g_state_switch_delay_counter = g_state_switch_max_delay;
    }
    else if (g_state == 1 && !g_state_switch_delay_counter && g_waterfall[g_waterfall_ptr][1] > 5.0) {
        //cout << "=============== State 2 ===============" << endl;
        g_state = 2;
        g_state_switch_delay_counter = g_state_switch_max_delay;
    }
    else if (g_state == 2 && !g_state_switch_delay_counter && g_waterfall[g_waterfall_ptr][1] > 3.0) {
        //cout << "=============== State 3 ===============" << endl;
        g_state = 3;
    }
    //===============================================================================
    // States
    if (g_state == 1) {
        glPushMatrix();
            glColor3f(1, 1, 1);
            const float target_radius = g_radius * g_RMS/g_RMS_threshold;
            g_radius0 = (target_radius + g_radius0) / 2;
            drawSolidCircle(0, 0, g_radius0);
        glPopMatrix();
    }
    else {
        if (g_state >= 2) {
            glPushMatrix();
                //===============================================================================
                // Rotation
                glRotatef(theAngle, 0, 0, 1);
                theAngle += .7;
                //===============================================================================
                // Spectrum Visualization
                glColor3f(1, 1, 1);
                glLineWidth(6.0);
                float min_radius_ratio = 0.08;
                for (int i = 0; i < g_waterfall_depth; ++i) {
                    int current_layer = (g_waterfall_ptr + g_waterfall_depth - i) % g_waterfall_depth;
                    float current_radius_ratio = 1 - (1 - min_radius_ratio) * i / g_waterfall_depth;
                    for (int j = 0; j < g_bufferSize; ++j){
                        const float darkness = g_waterfall[current_layer][j];
                        const float brightness = 1 - darkness;
                        glColor3f(brightness, brightness, brightness);
                        const float theta = (j == 0) ? PI/2 : PI/2 - 2*PI*(log10f(j*MY_SRATE/g_fftSize)-1)/(log10f(MY_SRATE/2)-1);
                        const float next_theta = PI/2 - 2*PI*(log10f((j+1)*MY_SRATE/g_fftSize)-1)/(log10f(MY_SRATE/2)-1);
                        float r, next_r;
                        if (next_theta >= -PI/2) {
                            const float rsin = g_radius*sinf(theta);
                            next_r = r = g_radius/2 * current_radius_ratio;
                        }
                        else if (theta >= -PI/2) {
                            const float next_rsin = g_radius*sinf(next_theta);
                            r = g_radius/2 * current_radius_ratio;
                            next_r = (next_rsin+sqrtf(next_rsin*next_rsin+3*g_radius*g_radius))/2 * current_radius_ratio;
                        }
                        else {
                            const float rsin = g_radius*sinf(theta);
                            const float next_rsin = g_radius*sinf(next_theta);
                            const float radius_sqr = g_radius*g_radius;
                            r = (rsin+sqrtf(rsin*rsin+3*radius_sqr))/2 * current_radius_ratio;
                            next_r = (next_rsin+sqrtf(next_rsin*next_rsin+3*radius_sqr))/2 * current_radius_ratio;
                        }
                        glBegin(GL_LINES);
                            glVertex2f(0+r*cosf(theta), -g_radius/2+r*sinf(theta));
                            glVertex2f(0+next_r*cosf(next_theta), -g_radius/2+next_r*sinf(next_theta));
                        glEnd();
                    }
                }
                //===============================================================================
                // Dark Side
                glColor3f(0, 0, 0);
                glLineWidth(3.0);
                for(float delta = 0; delta <= g_radius; delta += 0.0001) {
                    float arc_radius = sqrtf(g_radius*g_radius - g_radius*delta);
                    float delta_angle = atanf(sqrtf(g_radius*delta - delta*delta) / (g_radius - delta));
                    drawArc(0, 0, arc_radius, -PI/2.0+delta_angle, PI, 100);
                }
                //===============================================================================
                // Two Eyes
                const float g_radius_eye1 = g_eye_switch ? g_radius/8.6 : g_radius/9;
                const float g_radius_eye2 = !g_eye_switch ? g_radius/8.6 : g_radius/9;
                if (!g_eye_counter) {
                    g_eye_switch = !g_eye_switch;
                    g_eye_counter = 1;
                }
                else --g_eye_counter;
                drawSolidCircle(0, -g_radius/2, g_radius_eye1*1.1);
                glColor3f(1, 1, 1);
                drawSolidCircle(0, g_radius/2, g_radius_eye2);
                //===============================================================================
                // Waveform Visualization
                glColor3f(1, 1, 1);
                glLineWidth(1.2);
                glBegin(GL_LINE_STRIP);
                    for( int i = 0; i < g_bufferSize; ++i ){
                        const float theta = -PI/2 + i * PI / (g_bufferSize - 24);
                        // Slightly adjust the shape to cover the whole dark side
                        const float bass_ratio = 1.012;
                        float radius_ratio = bass_ratio - 3*fabs(buffer[i]);
                        const float rsin = g_radius*sinf(theta);
                        const float maxR = (-rsin+sqrtf(rsin*rsin+3*g_radius*g_radius))/2;
                        float R = radius_ratio * maxR;
                        if (theta <= -PI/3) {
                            const float minR = (-2*rsin+sqrtf(4*rsin*rsin-3*g_radius*g_radius)) / 2;
                            if (R < minR) R = minR;
                        }
                        else if (R < 0) R = -R < maxR ? -R : maxR;
                        glVertex2f(R*cosf(theta), g_radius/2 * bass_ratio + R*sinf(theta));
                    }
                glEnd();
                glLineWidth(1.0);
                glBegin(GL_LINE_STRIP);
                    for( int i = 0; i < g_bufferSize; ++i ){
                        const float theta = PI/2 + i * PI / g_bufferSize;
                        // Slightly adjust the shape to cover the whole dark side
                        const float bass_ratio = 1.012;
                        float radius_ratio = bass_ratio - 3*fabs(buffer[i]);
                        float R = radius_ratio * g_radius/2;
                        if (R < 0) R = -R < g_radius/2 ? -R : g_radius/2;
                        glVertex2f(R*cosf(theta), g_radius/2 + R*sinf(theta));
                    }
                glEnd();
            glPopMatrix();
        }
        if (g_state >= 3) {
            glPushMatrix();
                glRotatef(-theAngle, 0, 0, 1);
                glColor3f(1, 1, 1);
                drawBar(0, 3.7, 1.0, 0.16);
                drawBar(0, 3.35, 1.0, 0.16);
                drawBar(0, 3.0, 1.0, 0.16);

                glRotatef(45, 0, 0, 1);
                drawDashed(0, 3.7, 1.0, 0.16);
                drawBar(0, 3.35, 1.0, 0.16);
                drawBar(0, 3.0, 1.0, 0.16);

                glRotatef(45, 0, 0, 1);
                drawBar(0, 3.7, 1.0, 0.16);
                drawDashed(0, 3.35, 1.0, 0.16);
                drawBar(0, 3.0, 1.0, 0.16);

                glRotatef(45, 0, 0, 1);
                drawDashed(0, 3.7, 1.0, 0.16);
                drawDashed(0, 3.35, 1.0, 0.16);
                drawBar(0, 3.0, 1.0, 0.16);

                glRotatef(45, 0, 0, 1);
                drawDashed(0, 3.7, 1.0, 0.16);
                drawDashed(0, 3.35, 1.0, 0.16);
                drawDashed(0, 3.0, 1.0, 0.16);

                glRotatef(45, 0, 0, 1);
                drawBar(0, 3.7, 1.0, 0.16);
                drawDashed(0, 3.35, 1.0, 0.16);
                drawDashed(0, 3.0, 1.0, 0.16);

                glRotatef(45, 0, 0, 1);
                drawDashed(0, 3.7, 1.0, 0.16);
                drawBar(0, 3.35, 1.0, 0.16);
                drawDashed(0, 3.0, 1.0, 0.16);

                glRotatef(45, 0, 0, 1);
                drawBar(0, 3.7, 1.0, 0.16);
                drawBar(0, 3.35, 1.0, 0.16);
                drawDashed(0, 3.0, 1.0, 0.16);
            glPopMatrix();
        }
    }
    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
    // */

    //===============================================================================
    //===============================================================================
    /* Normal Visualization Test
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    // line width
    glLineWidth( 1.0 );
    //===============================================================================
    // Waveform Visualization
    glColor3f( .5, 1, .5 );
    apply_window((float*)g_buffer, g_window, g_bufferSize);
    // start primitive
    glBegin( GL_LINE_STRIP );
        // loop over buffer
        GLfloat width = 10;
        for( int i = 0; i < g_bufferSize; ++i ){
            glVertex2f(-width/2 + i*width/g_bufferSize, -2 + 3*g_buffer[i]);
        }
    // end primitive
    glEnd();
    //===============================================================================
    // Spectrum Waterfall Visualization
    glColor3f(0, 0.5, 1);
    SAMPLE buffer[g_fftSize];
    memset(buffer, 0, g_fftSize * sizeof(SAMPLE));
    memcpy(buffer, g_buffer, g_bufferSize * sizeof(SAMPLE));
    rfft((float *)buffer, g_bufferSize, FFT_FORWARD);
    complex* cbuf = (complex *)buffer;
    g_waterfall_ptr = (g_waterfall_ptr + 1) % g_waterfall_depth;
    for (int i = 0; i < g_bufferSize; ++i) {
        g_waterfall[g_waterfall_ptr][i] = sqrtf(200*cmp_abs(cbuf[i]));
    }
    float max_width = 10;
    float min_horizon = -0.5;
    float width_step = -0.2;
    float horizon_step = 0.05;
    float waterfall_decay = 0.945;
    for (int i = 0; i < g_waterfall_depth; ++i) {
        int current_layer = (g_waterfall_ptr + g_waterfall_depth - i) % g_waterfall_depth;
        float current_gain = 1.0 - (float)i / g_waterfall_depth;
        glColor3f(0, 0.5*current_gain, 1*current_gain);
        float current_width = max_width + i*width_step;
        float current_horizon = min_horizon + i*horizon_step;
        glBegin(GL_LINE_STRIP);
            for (int j = 0; j < g_bufferSize; ++j){
                glVertex2f(-current_width/2 + j*current_width/g_bufferSize, current_horizon + g_waterfall[current_layer][j]);
            }
        glEnd();
    }
    //===============================================================================
    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
    // */
    //===============================================================================
    //===============================================================================
}
